<?php
/*
Plugin Name: Biuro Feedbacks
Description: Biuro Feedbacks plugin
Author: Biuro
Version: 1.0
Author URI: https://www.biuro.lt/
*/


class Biuro_Feedbacks extends WP_Widget {
    // Main constructor
    public function __construct() {
        parent::__construct(
            'biuro-feedbacks',
            __( 'Biuro Feedbacks | Biuro', 'biuro' ),
            array(
                'customize_selective_refresh' => true,
            )
        );
    }

    // The widget form (for the backend )
    public function form( $instance ) {
        // Set widget defaults
        $defaults = array(
            'title' => '',
            'name1'    => '',
            'img1'    => '',
            'feedback1' => '',
            'name2'    => '',
            'img2'    => '',
            'feedback2' => '',
            'name3'    => '',
            'img3'    => '',
            'feedback3' => '',
            'name4'    => '',
            'img4'    => '',
            'feedback4' => '',
            'name5'    => '',
            'img5'    => '',
            'feedback5' => '',
            'name6'    => '',
            'img6'    => '',
            'feedback6' => '',
            'name7'    => '',
            'img7'    => '',
            'feedback7' => '',
            'name8'    => '',
            'img8'    => '',
            'feedback8' => '',
        );


        // Parse current settings with defaults
        extract( wp_parse_args( ( array ) $instance, $defaults ) ); ?>

        <p>
            <label for="<?php echo esc_attr( $this->get_field_id( 'title' ) ); ?>"><?php _e( 'Title', 'biuro' ); ?><?php echo ' ' . $i . ':';  ?></label>
            <input class="widefat" id="<?php echo esc_attr( $this->get_field_id( 'title' ) ); ?>" name="<?php echo esc_attr( $this->get_field_name( 'title' ) ); ?>" type="text" value="<?php echo esc_attr( ${'title'} ); ?>" />
        </p>
<?php
        for ($i = 1; $i <= 8; $i++):
?>
        <p>
            <label for="<?php echo esc_attr( $this->get_field_id( 'name' . $i ) ); ?>"><?php _e( 'Name', 'biuro' ); ?><?php echo ' ' . $i . ':';  ?></label>
            <input class="widefat" id="<?php echo esc_attr( $this->get_field_id( 'name' . $i ) ); ?>" name="<?php echo esc_attr( $this->get_field_name( 'name' . $i ) ); ?>" type="text" value="<?php echo esc_attr( ${'name' . $i} ); ?>" />
        </p>
        <p>
            <label for="<?php echo esc_attr( $this->get_field_id( 'img' . $i ) ); ?>"><?php _e( 'Image', 'biuro' ); ?><?php echo ' ' . $i . ':';  ?></label>
            <input class="widefat" id="<?php echo esc_attr( $this->get_field_id( 'img' . $i ) ); ?>" name="<?php echo esc_attr( $this->get_field_name( 'img' . $i ) ); ?>" type="text" value="<?php echo esc_attr( ${'img' . $i} ); ?>" />
        </p>
        <p>
            <label for="<?php echo esc_attr( $this->get_field_id( 'feedback' . $i ) ); ?>"><?php _e( 'Feedback', 'biuro' ); ?><?php echo ' ' . $i . ':';  ?></label>
            <textarea class="widefat" id="<?php echo esc_attr( $this->get_field_id( 'feedback' . $i ) ); ?>" name="<?php echo esc_attr( $this->get_field_name( 'feedback' . $i ) ); ?>" rows="4"><?php echo wp_kses_post( ${'feedback' . $i} ); ?></textarea>
        </p>
<?php
        endfor;
 ?>

    <?php }

    // Update widget settings
    public function update( $new_instance, $old_instance ) {
        $instance = $old_instance;

        $instance['title']    = isset( $new_instance['title'] ) ? wp_strip_all_tags( $new_instance['title'] ) : '';

        for ($i = 1; $i <= 8; $i++):
            $instance['name' . $i]    = isset( $new_instance['name' . $i] ) ? wp_strip_all_tags( $new_instance['name' . $i] ) : '';
            $instance['img' . $i]    = isset( $new_instance['img' . $i] ) ? wp_strip_all_tags( $new_instance['img' . $i] ) : '';
            $instance['feedback' . $i] = isset( $new_instance['feedback' . $i] ) ? wp_kses_post( $new_instance['feedback' . $i] ) : '';
        endfor;

        return $instance;
    }

    // Display the widget
    public function widget( $args, $instance ) {
        extract( $args );

        $title = isset( $instance['title'] ) ?$instance['title'] : '';

        for ($i = 1; $i <= 8; $i++):
            ${'name' . $i} = isset( $instance['name' . $i] ) ?$instance['name' . $i] : '';
            ${'img' . $i} = isset( $instance['img' . $i] ) ?$instance['img' . $i] : '';
            ${'feedback' . $i} = isset( $instance['feedback' . $i] ) ?$instance['feedback' . $i] : '';
        endfor;

        // WordPress core before_widget hook (always include )
        echo $before_widget;
        // Display the widget
        echo '<div class="c-feedbacks">';

            if ( $title != "" ) {
                echo '<h2 class="c-feedbacks--heading">' . $title . '</h2>';
            }

            echo '<div class="c-feedbacks--inner">';
                echo '<div class="swiper-wrapper">';

                for ($i = 1; $i <= 8; $i++):

                    if ( ${'name' . $i} != "" && ${'feedback' . $i} != ""):
                        echo '<div class="swiper-slide">';
                            echo '<div class="c-feedbacks--section">';
                                echo '<div class="c-feedbacks--section-inner">';

                                if ( ${'img' . $i} != "" ) {
                                    $img = preg_replace('/\.png$/i', '', ${'img' . $i});

                                    echo '<picture>';
                                        echo '<source srcset="/wp-content/themes/biuro/i/feedbacks/' . $img . '.webp" type="image/webp">';
                                        echo '<img loading="lazy" src="/wp-content/themes/biuro/i/feedbacks/' . $img . '.png" width="85px" height="83px" alt="' . ${'name' . $i} . '" class="c-feedbacks--img">';
                                    echo '</picture>';
                                }

                                if ( ${'feedback' . $i} != "" ) {
                                    echo '<div class="c-feedbacks--feedback">' . ${'feedback' . $i} . '</div>';
                                }

                                if ( ${'name' . $i} != "" ) {
                                    echo '<div class="c-feedbacks--name">' . ${'name' . $i} . '</div>';
                                }

                                echo '</div>';
                            echo '</div>';
                        echo '</div>';
                    endif;

                endfor;

                echo '</div>';
                echo '<div class="swiper-pagination"></div>';
            echo '</div>';
        echo '</div>';

        // WordPress core after_widget hook (always include )
        echo $after_widget;
    }
}


// register Biuro_Feedbacks
add_action( 'widgets_init', function(){
    register_widget( 'Biuro_Feedbacks' );
});

